/*********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <process.h>
#include <sys/types.h>
#include <math.h>
#include <conio.h>
#include <process.h>
#include <time.h>
#include <dos.h>
//#include <VSGEX2.h>
#include <VSGV8.h>
 

/*********************************************************************/
#define DRND			((double) rand() / (double) RAND_MAX)
#define APEWIDTH		2.7
#define MAXLONG     0x7fffffff  

#define ON				1
#define OFF				0
#define YES				0
#define NO				1
#define FALSE			0
#define TRUE			1
#define FULL			0
#define PARTIAL		1
#define HEADS			1
#define TAILS			0

#define CORRECT_SOUND (200, 400)
#define INCORRECT_SOUND (100, 100)
#define TONE_LENGTH	250
#define LOW_TONE	200
#define HIGH_TONE 400
#define LONG_TONE	300
#define SHORT_TONE 200

#define BUTTON_NONE  1
#define A_UP   		2
#define B_UP   		4
#define C_UP   		8
#define A_DOWN			16
#define B_DOWN			32
#define C_DOWN			64
#define EQUAL_TO		128

#define LEFTARROW		75
#define RIGHTARROW	77
#define UPARROW		72
#define DOWNARROW		80
#define ESCAPE			27
#define EXTENDED		224

#define OUTFILE		"output.ape"
#define DELAY			500


/*********************************************************************/

typedef char boolean;
typedef int analtype;

/*********************************************************************/

static void		(*initf)(char *[]) = NULL;
static void		(*runf)(int, double, char *[]) = NULL;
static void		(*endf)(char *[]) = NULL;

static int		(*LoopButton)(int) = NULL;
static void		(*ClearButton)(void) = NULL;
static void		(*ResponseSound)(int, int) = NULL;

static char		**onestimargs = NULL;
static char		**startargs = NULL;
static char		**endargs = NULL;
static char		*apeoutput = NULL;
static char		*apeinput = NULL;
static char		***stimargs = NULL;

static int		apelevels = 32;
static int		dragtrials = 16;
static int		apetrials = 64;
static int		aperuns  = 1;

static int		no_levels;
static int		*trial, *stimulus, response;
static int		run = 0, prev_run, trial_sum;
static int		expected = BUTTON_NONE;
static int		grandnumpres = 0;

static double		*apebias;
static double		defbias = 0.0;
static double		*apeoffset;
static double		defoffset = 0.5;
static double		*apegain;
static double		defgain = 0.01;
static double		*ape_sd, *ape_ed50, *Csd, *Ced50;
static double		anal_sd, sderr, anal_ed50, ed50err;

static double		grandtotal = 0.0;
static double		lgrandtotal = 0.0;

static boolean *feedback;
static boolean	overflow;
static boolean	ctadapt = TRUE;
static boolean	apeinvert = NO;
static boolean	apequiet = NO;
static boolean	printape = NO;

/*********************************************************************/


static struct apestruct {
	int *stims;
	int *resps;
} srlist;

static struct analstruct {
	int    *stim;
	int    *test;
	int    *resp;
	double *prop;
	double *prob;
	double *weight;
} srtable;

static struct infilestuff {
	char *ptr;
	int lines;
	int args;
} apein;

static struct ptrstuff {
	char ***p;
	int *stm;
	int *apl;
} inptr;


CBBOXRECORD CB6Box;  // Identify CB3 response box



/*********************************************************************/
int ApeMain(int argc, char *argv[], void *initfunc, void *runfunc, void *endfunc);
/*********************************************************************/

static double Probit(double x);
static double InvProbit(double x);
static void ApeEngine(int trial, int *stim, int response);
static void ApeAdapt(int trial);
static void ApeAnalyse(analtype analysis, int start, int finish, FILE *out);
static void ApeRegress(double *a,double *aerr,double *b,double *berr);
static int ButtonResponse(int button_v);
static int LoopButtonbox(int invert);
static void ApeGetFirstTrial(void);
static void ApeGetNextTrial(void);
static void ApeAnalyser(void);
static void ApeRunTrials(void);
static int ApeRunsLeft(int *ptr, int limit, int len);
static int ApeNextRun(void);

/*********************************************************************/

static int StrAndPtrs(char *infile);
static int CreateString(char *infile);
static int ApeInCount(FILE *in);
static int ApeSuckIn(FILE *in, char *str);
static int NumArgs(char *strptr);
static void **ApeSetPointers(int numlines, char *str);
static void ReDoString(char *str);
static void PrintArgs(void);
static int CheckArgs(int argc, char *argv[], char *str);
static int GetNumRuns(int lines);
static void ApeInitArgs(int argc, char *argv[]);
static void ApeInitRuns(void);
static char **GetStimPtr(int i);
static void ApeDoArgs(int argc, char **argv);
static void ApeInit(int argc, char *argv[]);
static void ApeInitMem(void);
static void ApeClearMem(void);
static void *ApeCheckMem(void *ptr);
static void ApeInitVals(char *id, double defval, double *doubleptr);
static char *dtoa(double num, char *str, int length);
static char *DoApeParameters(double dnum, boolean invt, char *str, int length);
static int DisplayStimulus(double parameter, double baseval);
static void PrintParameters(void);
static void PrintArgv(char *argv[], FILE *out);
static void PrintInfo(void);
static void SetMemToZero(void);
static void SetupResponse(void);
static void ResponsePC(int actual, int expected);
static void RunCommandCom(void);
static void do_sound(unsigned int first, unsigned int second);
static void UseFeedback(void);
static int flip_coin(void);
static int dtoi(double num);
static void ClearKeyboard(void);
static void LogVals(void);
static void PrintResults(FILE *in);
static double sq(double num);
static double ChiSquared(double a, double b);
static double Convert(double num);
static int GetStimVal(double offset, double sd);
static void AddBias(void);
static void do_wrong_sound(void);
static void do_correct_sound(void);
void wrong_buzzer(void);
void correct_buzzer(void);

/*********************************************************************/

int ApeMain(int argc, char *argv[], void *initfunc, void *runfunc, void *endfunc)
{
	if (argc == 1 || runfunc == NULL) {
		PrintInfo();
		return 1;
	}
	else {

		initf = (void (*)(char *[])) initfunc;
		runf = (void (*) (int, double, char *[])) runfunc;
		endf = (void (*) (char *[])) endfunc;
		ApeInit(argc, argv);
		if (apeinput == NULL && onestimargs == NULL) {
			PrintInfo();
			return 1;
		}
		else {
			srand(time(NULL));
			if (apequiet == NO) PrintParameters();
			SetupResponse();
			if (initf != NULL)
				(*initf)(startargs);
			ApeRunTrials();
			if (endf != NULL)
				(*endf)(endargs);
			ApeAnalyser();
			ApeClearMem();
			return 0;
		}
	}

}

double Probit(double x)
{
	double lx,l1,l2,l3;

	if (x < 0.001)  x = 0.001;
	if (x > 0.999)  x = 0.999;
	lx = (x > 0.5) ? 1.0 - x : x;

	l1 = log(1.0/lx/lx);
	l2 = sqrt(l1);
	l3 = (2.515517+0.802853*l2+0.010328*l1);
	l3 = l3/(1.0+1.432788*l2+0.189269*l1+0.001308*l1*l2);
	l3 = l2-l3;

	return (x < 0.5) ? -l3 : l3;

}

double InvProbit(double x)
{
	double l1,l2,l3;

	l1 = fabs(x);
	l1 = 1.0/(1.0+0.231642*l1);
	l2 = 0.398942*exp(-x*x/2.0);
	l3 = (((1.330274*l1-1.821256)*l1+1.781478)*l1-0.356564)*l1+0.319382;
	l3 = 1.0-l1*l2*l3;
	if (x < 0.0)  l3 = 1.0-l3;
	if (l3 < 0.001)  l3 = 0.001;
	if (l3 > 0.999)  l3 = 0.999;
	return(l3);
}

void ApeEngine(int trial, int *stim, int response)
{
	if (trial >= 0) {
		*(srlist.stims + (trial * aperuns) + run) = *stim;
		*(srlist.resps + (trial * aperuns) + run) = response;
	}
	ApeAdapt(trial);
	if (trial < dragtrials)
		*stim =
			(flip_coin() == HEADS) ? dtoi(*(ape_sd + run)) : -1 * dtoi(*(ape_sd + run));
	else {
		*stim =
			dtoi((((DRND * APEWIDTH) - (APEWIDTH / 2.0)) * *(ape_sd + run)) + *(ape_ed50 + run));
	}
}

int dtoi(double num)
{
	return (num < 0.0) ? (int) (num - 0.5) : (int) (num + 0.5);
}

void ApeAdapt(int trial)
{
	int   sump,i;
	FILE *out = stderr;

	if (trial >= dragtrials) {
		if (trial > dragtrials*2)
			ApeAnalyse(PARTIAL,trial-dragtrials*2,trial, out);
		else
			ApeAnalyse(PARTIAL,0,trial, out);
		anal_sd = fabs(anal_sd);
		if (anal_sd > 10.0){
			anal_ed50 = 0.0;
			anal_sd = 10.0;
		}
		*(ape_ed50 + run) += *(Ced50 + run) * (anal_ed50 - *(ape_ed50 + run));
		*(ape_sd + run)   *= exp( *(Csd + run) * log(anal_sd / *(ape_sd + run)));
		*(ape_sd + run) = fabs(*(ape_sd + run));

		*(Csd + run) = fabs(log(anal_sd / *(ape_sd + run)));
		*(Ced50 + run) = fabs(anal_ed50 - *(ape_ed50 + run)) / *(ape_sd + run);

		if (*(Ced50 + run) > 0.5)
			*(Ced50 + run) = 0.5;
		if (*(Csd + run)   > 0.5)
			*(Csd + run)   = 0.5;
	}
	else {
		if (trial < 2) {
			anal_sd = apelevels/4.0;
		}
		else {
			sump = 0;
			for (i = 0; i < trial; i++) {
				if  (((*(srlist.stims + (i * aperuns) + run) > 0) &&
					(*(srlist.resps + (i * aperuns) + run) ==1)) ||
					((*(srlist.stims + (i * aperuns) + run) < 0) &&
					(*(srlist.resps + (i * aperuns) + run) ==0)))
						sump = sump + 1;
			}
			if (sump >= trial/2) {
				if  (((*(srlist.stims + (trial * aperuns) + run) > 0) &&
					(*(srlist.resps + (trial * aperuns) + run) ==1)) ||
					((*(srlist.stims + (trial * aperuns) + run) < 0) &&
					(*(srlist.resps + (trial * aperuns) + run) ==0)))
						anal_sd = *(ape_sd + run) / sqrt(2.0);
				else
					anal_sd = *(ape_sd + run) * sqrt(2.0);
			}
			else {
				if  (((*(srlist.stims + (trial * aperuns)  + run) > 0) &&
					(*(srlist.resps + (trial * aperuns) + run) ==0)) ||
					((*(srlist.stims + (trial * aperuns) + run) < 0) &&
					(*(srlist.resps + (trial * aperuns) + run) ==1)))
						anal_sd = *(ape_sd + run) / sqrt(2.0);
					else
						anal_sd = *(ape_sd + run) * sqrt(2.0);
			}
			if (anal_sd < 1.0)  anal_sd = 1.0;
		}
		anal_ed50        = 0.0;
		*(ape_sd + run)      = anal_sd;
		*(ape_ed50 + run)    = anal_ed50;
		*(Ced50 + run)       = 0.5;
		*(Csd + run)         = 0.7;
	}
	/* CB 7/3/95: *(ape_ed50 + run) used to equal zero */
	/* 10/3/95 set back to zero. ape_ed50 seems to be  */
	/* the difference between current estimate of PSE  */
	/* and apeoffset.                                  */
	if (ctadapt == FALSE)
		*(ape_ed50 + run) = 0.0;
}

void ApeAnalyse(analtype analysis, int start, int finish, FILE *out)
{
	double a,b,az,bz ;
	double aerr,berr;
	int    i1,i2;
	int    iterations,cycles;
	double exprob,exprtn,z;
	boolean    found;
	int    lowest;

	iterations = (analysis == FULL) ? 40 : 0;

/* start by making a list of stimuli */

	no_levels = 0;
	for (i1 = start; i1 < finish; i1++) {
		found = FALSE;
		for (i2 = 0; i2 < no_levels; i2++)
			if (*(srtable.stim +i2) == *(srlist.stims + (i1 * aperuns) + run))
				found = TRUE;
		if (found != TRUE)
				*(srtable.stim + no_levels++) = *(srlist.stims + (i1 * aperuns) + run);
	}

 /* sort into order */

	for (i1 = 0; i1 < no_levels;i1++) {
		lowest = i1;
		for (i2 = i1+1; i2 < no_levels;i2++)
			if (*(srtable.stim + i2) < *(srtable.stim + lowest))
				lowest = i2;
			*(srtable.stim + no_levels)  = *(srtable.stim + i1);
			*(srtable.stim + i1)         = *(srtable.stim + lowest);
			*(srtable.stim + lowest)     = *(srtable.stim + no_levels);
			*(srtable.test + i1)         = 0;
			*(srtable.resp + i1)         = 0;
	}

/* then sort into a list of stimulus/response */

	for (i1 = start; i1 < finish; i1++) {
		for (i2 = 0; *(srlist.stims + (i1 * aperuns) + run) > *(srtable.stim + i2); i2++);
			*(srtable.test + i2) += 1;
		*(srtable.resp + i2) += *(srlist.resps + (i1 * aperuns) + run);
	}

 /* then calculate proportions, probits & weights */

	for (i1 = 0;i1 < no_levels;i1++) {
		*(srtable.prop + i1) = (double) *(srtable.resp + i1)/(double) *(srtable.test + i1);
		*(srtable.prob + i1) = Probit(*(srtable.prop + i1));
		*(srtable.weight + i1) = 1.0;
	}

 /* first (unweighted) regression */


	ApeRegress(&a,&aerr,&b,&berr);
	if (overflow == TRUE) {
		b = 0.0;
		a = 0.0;
	}
	cycles = 0;
	az = a + 1.0;
	bz = b + 1.0;

	while ( (cycles < iterations)
		&& (fabs(a-az) > 0.001) && (fabs(b-bz) > 0.001)
		&& (overflow == FALSE)) {

	 /* calculate working probits & weights */

		for (i1 = 0; i1 < no_levels; i1++) {

			exprob = (double) *(srtable.stim + i1) * b + a;
			exprtn = InvProbit(exprob);
			z      = exp(-exprob*exprob/2.0) / 2.506628;

			if ((z != 0.0) && (exprtn != 0.0) ) {
				*(srtable.prob + i1) = exprob+(*(srtable.prop + i1)-exprtn)/z;
				*(srtable.weight + i1) = (z*z/(exprtn*(1.0-exprtn)))
					* *(srtable.test + i1);
			}

			else *(srtable.weight + i1) = 0.0;
		}

		az = a;
		bz = b;
		ApeRegress(&a,&aerr,&b,&berr);
		cycles = cycles +1;
		if (overflow == TRUE) {
			fprintf(out, "  **OVERFLOW**\n");
		}
	}

	if (fabs(b) <= 0.0001)
		b = 0.0001;


	anal_sd   = 1.0/b;
	sderr     = fabs(berr/b/b);
	anal_ed50 = -a/b;
	ed50err   = fabs(aerr/b);

	if (analysis == FULL) {
		fprintf(out, "offset = %f\ngain   = %f\nbias   = %f\n",
			*(apeoffset + run) - (*(apegain + run) * *(apebias + run)),
			*(apegain + run), *(apebias + run));
		PrintArgv(*(stimargs + run), out);
		fprintf(out, "\n");

		anal_sd   = anal_sd * *(apegain + run);
		sderr     = sderr * *(apegain + run);
		anal_ed50 = anal_ed50 * *(apegain + run) + *(apeoffset + run);
		ed50err   = ed50err * *(apegain + run);

		PrintResults(out);
	}
}

void ApeRegress(double *a,double *aerr,double *b,double *berr)
{
	double snw = 0.0,snwx = 0.0,snwy = 0.0,snwxy = 0.0,snwxx = 0.0,snwyy = 0.0;
	double w,x,y;
	int    i;


	for (i = 0; i < no_levels; i++) {
		w = *(srtable.weight + i);
		x = (double) *(srtable.stim + i);
		y = *(srtable.prob + i);
		snw   = snw   + w;
		snwx  = snwx  + w*x;
		snwy  = snwy  + w*y;
		snwxx = snwxx + w*x*x;
		snwyy = snwyy + w*y*y;
		snwxy = snwxy + w*x*y;
	}

	if  ((snwxx*snw - snwx*snwx != 0.0) && (snw != 0.0)) {
		*b    = (snwxy*snw - snwx*snwy)/(snwxx*snw - snwx*snwx);
		*berr = sqrt(snw/(snwxx*snw - snwx*snwx));
		*a    = snwy/snw - *b*snwx/snw;
		*aerr = sqrt(1.0/snw);

		overflow = FALSE;
	}
	else
		overflow = TRUE;
}

void PrintResults(FILE *out)
{
#define CONF_95 1.96

	double snw, snwx, snwy, snwxy, snwxx, snwyy;
	double s_xx, s_xy, s_yy;
	double w, x_bar, y_bar, g, m, x, y;
	double a, b, aerr, berr;
	double tmp1, tmp2;
	double std_err;
	double lstim;
	double total, logtotal;
	int    i, numpres;

	snw = snwx = snwy = snwxy = snwxx = snwyy = 0.0;
	total = logtotal = 0.0;
	numpres = 0;

	fprintf(out, "\n      z         x         n    r       p        NED        Y\n");

	for (i = 0; i < no_levels; i++) {
		lstim = ((((double) *(srtable.stim + i)) * *(apegain + run)) + *(apeoffset + run));
		fprintf(out, "%9.3f %9.3f %7d %4d %9.3f %9.3f %9.3f\n", pow(10, lstim), lstim, *(srtable.test + i), *(srtable.resp + i), (double) *(srtable.prop + i), (double) Probit(*(srtable.prop + i)), (double) *(srtable.prob + i));
		if (srtable.test[i] > 0) {
			w = *(srtable.weight + i);
			x = Convert(*(srtable.stim + i));
			y = *(srtable.prob + i);
			snw   = snw   + w;
			snwx  = snwx  + w*x;
			snwy  = snwy  + w*y;
			snwxx = snwxx + w*x*x;
			snwyy = snwyy + w*y*y;
			snwxy = snwxy + w*x*y;
			logtotal += lstim * ((double) *(srtable.test + i));
			total += pow(10, lstim) * *(srtable.test + i);
			numpres += *(srtable.test + i);
		}
	}

	b = (((snwxy * snw) - (snwx * snwy)) / ((snwxx * snw) - (snwx * snwx)));
	berr = sqrt(snw / ((snwxx * snw) - (snwx * snwx)));
	a = (snwy / snw) - (b * snwx / snw);
	aerr = sqrt(1.0 / snw);

	x_bar = snwx / snw;
	y_bar = snwy / snw;

	s_xx = snwxx - (sq(snwx) / snw);
	s_xy = snwxy - ((snwx * snwy) / snw);
	s_yy = snwyy - (sq(snwy) / snw);

	g = sq(CONF_95) / (sq(b) * s_xx);
	m = (-y_bar / b) + x_bar;

	/* print 95% confidence limits - see Finney, 1971 p79 */
	tmp1 = m + ((g / (1.0 - g)) * (m - x_bar));
	tmp2 = CONF_95 / (b * (1.0 - g)) * sqrt(((1.0 - g) / snw) + (sq(m - x_bar) / s_xx));

	/* returns the standard error as given by Finney, 1971 p34 eqn 3.12 */
	std_err = (1.0 / b) * sqrt((1.0 / snw) + (sq(m - (snwx / snw))/(snwxx - (sq(snwx) / snw))));
	std_err *= pow(10.0, m) * log(10.0);

	fprintf(out, "\nregression: a = %f  b = %f\n\n", a, b);

	fprintf(out, "log mean       = %f\n", (double) (logtotal / numpres));
	fprintf(out, "mean           = %f\n", (double) (total / numpres));
	fprintf(out, "log ED50       = %f (%f)\n", m, (double) fabs(aerr/b));
	fprintf(out, "ED50           = %f (%f - %f)\n", pow(10, m), pow(10, tmp1 + tmp2), pow(10, tmp1 - tmp2));
	fprintf(out, "standard error = %f\n", (double) fabs(std_err));
	fprintf(out, "log SD         = %f (%f)\n", (double) fabs(1.0 / b), (double) fabs(berr/sq(b)));
	fprintf(out, "chi squared    = %f (%d)\n\n", (double) ChiSquared(a, b), no_levels - 2);

	grandnumpres += numpres;
	grandtotal += total;
	lgrandtotal += logtotal;

#undef CONF_95
}

double Convert(double num)
{
	return ((num * *(apegain + run)) + *(apeoffset + run));
}

double sq(double num)
{
	return (num * num);
}

double ChiSquared(double a, double b)
{
	double P, chi2;
	int i;

	for (i = 0, chi2 = 0.0; i < no_levels; ++i) {
		P = InvProbit(a + (b * Convert(srtable.stim[i])));
		if (P != 0.0 && P != 1.0 && srtable.test[i] != 0.0)
			chi2 += sq(srtable.resp[i] - (srtable.test[i] * P)) / (srtable.test[i] * P * (1.0 - P));
	}
	return chi2;
}
		


void ApeRunTrials(void)
{
	int button = BUTTON_NONE;


	while (!ButtonResponse(button) && (ApeRunsLeft(trial, apetrials, aperuns) > 0)) {
		button = (*LoopButton)(
						DisplayStimulus(
							*(apeoffset + run) + (*(apegain + run) * *(stimulus + run)),
							*(apeoffset + run)
						)
					);
	}
}

int ButtonResponse(int button_v)
{
	switch (button_v) {
	case A_UP:
	case C_UP:
		if (*(feedback + run) == YES)
			(*ResponseSound)(button_v, expected);
		response = (button_v == C_UP) ? 0 : 1;
		ApeGetNextTrial();
		if (printape == YES) printf("ape value: %d\n", *(stimulus + run));
		break;

	case BUTTON_NONE:
		ApeGetFirstTrial();
		if (printape == YES) printf("ape value: %d\n", *(stimulus + run));
		break;
	case B_UP:
		break;
	}
	return 0;
}


int LoopButtonbox(int invert)
{
	int retval = BUTTON_NONE;
	long response_time = 1000 * 1000L, c;

	cbboxFlush(); // clear CB3 box
	vsgResetTimer();

  do
  {
    if (cbboxCheck((CBBOXRECORD*)&CB6Box)!=respEMPTY)
    {
      if (CB6Box.switchB == respUP)
      {
     //   cbboxBuzzer(respSEC01, respTONE1);
		retval = B_UP;
      }
      else if (CB6Box.switchA == respUP)
      {
     //   cbboxBuzzer(respSEC01, respTONE1);
		retval = (invert == 1) ? C_UP : A_UP;
      }
      else if (CB6Box.switchC == respUP)
      {
     //   cbboxBuzzer(respSEC01, respTONE1);
		retval = (invert == 1) ? A_UP : C_UP;
      }
	}
	  else if (vsgGetTimer() >= response_time){
		cbboxBuzzer(respSEC01, respTONE10);
		retval = B_UP;
		}
	  else if(kbhit()) {
		  if((c = getch()) == ESCAPE)
			  retval = ESCAPE;
		}
  }
  while (retval != A_UP && retval != C_UP && retval != B_UP && retval != ESCAPE);
 

  if (retval == ESCAPE){
				if (endf != NULL)
					(*endf)(endargs);
				cbboxClose();
				exit(1);
			}

  return retval;
}


void ClearKeyboard(void)
{
	while (kbhit())
		getch();

	cbboxFlush(); // clear CB3 box

}


void ResponsePC(int actual, int expected)
{
	if (actual == expected)
		do_correct_sound();
	else if (expected == EQUAL_TO)
		(flip_coin() == HEADS) ? do_sound CORRECT_SOUND : do_sound INCORRECT_SOUND;
	else
		do_wrong_sound();
}


int flip_coin(void)
{
	return (rand() > (RAND_MAX / 2.0)) ? HEADS : TAILS;
}

void do_sound(unsigned int first, unsigned int second)
{
}

void do_correct_sound(void)
{
}

void do_wrong_sound(void)
{
}

void ApeGetFirstTrial(void)
{
	trial_sum = 0;
	for (run = 0; run < aperuns; run++) {
		*(trial + run) = 0;
		response = 0;
		ApeEngine(*(trial + run), stimulus + run, response);   /* get first trial */
	}
	run = (int) (((double) rand() / RAND_MAX) * aperuns);
}

void ApeGetNextTrial(void)
{
	ApeEngine(*(trial + run), stimulus + run, response);   /*  get next trial */
	++trial_sum;
	++*(trial + run);
	prev_run = run;
	run = ApeNextRun();
}

int ApeRunsLeft(int *ptr, int limit, int len)
{
	int i, retval = 0;
	for (i = 0; i < len; ++i) {
		if (*(ptr + i) < limit)
			++retval;
	}
	return retval;
}

int ApeNextRun(void)
{
	int left, i, n, rndval;
	left = ApeRunsLeft(trial, apetrials, aperuns);
	rndval = (int) (((double) rand()/ RAND_MAX) * left);
	for (i = n = 0; n <= rndval && i < aperuns; ++i) {
		if (*(trial + i) < apetrials)
			++n;
	}
	return (i - 1);
}

void ApeAnalyser(void)
{
	FILE *out;
	time_t now;

	if (apeoutput != NULL) {
		if ((out = fopen(apeoutput, "at")) == NULL) {
				fprintf(stderr, "Cannot open requested output file - will try to open default output file\n\n");
				if ((out = fopen(OUTFILE, "at")) == NULL) {
					fprintf(stderr, "Cannot open default output file - will write to stdout\n\n");
					out = stdout;
				}
		}
	}
	else {
		if ((out = fopen(OUTFILE, "at")) == NULL) {
			fprintf(stderr, "Cannot open default output file\nwill write to stdout\n");
			out = stdout;
		}
	}

	if (out != stdout) {
		now = time(NULL);
		fprintf(out, "************************************************************\n");
		fprintf(out, "%s\n", asctime(localtime(&now)));
	}

	for (run = 0;run < aperuns; run++) {
		ApeAnalyse(FULL,0,apetrials, out);
		fprintf(out, "************************************************************\n");
	}

	if (aperuns > 1) {
		fprintf(out, "\ngrand mean     = %f\n", grandtotal / (double) grandnumpres);
		fprintf(out, "log grand mean = %f\n", lgrandtotal / (double) grandnumpres);
	}

	fclose(out);
}

/*********************************************************************/

int StrAndPtrs(char *infile)
{
	int i;

	apein.lines = CreateString(infile);
	inptr.apl = (int *) ApeCheckMem(malloc(sizeof(int) * apein.lines));
	inptr.stm = (int *) ApeCheckMem(malloc(sizeof(int) * apein.lines));
	for (i = 0; i < apein.lines ; ++i)
		*(inptr.apl + i) = *(inptr.stm + i) = 0;

	apein.args = NumArgs(apein.ptr);

	inptr.p = (char ***) ApeCheckMem(malloc(sizeof(void *) * (apein.args + (apein.lines * 2) + 1)));
	ApeSetPointers(apein.lines, apein.ptr);

	ReDoString(apein.ptr);

	return GetNumRuns(apein.lines);
}

int CreateString(char *infile)
{
	FILE *in;
	int lines;

	if ((in = fopen(infile, "rt")) == NULL) {
		fprintf(stderr, "Cannot open input file: %s\n", infile);
		exit(1);
	}
	apein.ptr = (char *) ApeCheckMem(malloc(ApeInCount(in)));
	lines = ApeSuckIn(in, apein.ptr);
	fclose(in);
	return lines;
}


int ApeInCount(FILE *in)
{
	int outprev = '\n', c, post, charnum = 0;
	char count = YES;

	fseek(in, 0L, SEEK_SET);

	c = fgetc(in);
	post = fgetc(in);
	do {
		if (c == '#')
			count = NO;
		else if (c == '\n')
			count = YES;
		if (count == YES) {
			if (outprev == ' ' && c == ' ')
				;
			else if (outprev == '\n' && c == '\n')
				;
			else if (outprev == '\n' && c == ' ')
				;
			else if (c == ' ' && post == '#')
				;
			else if (c == ' ' && post == '\n')
				;
			else {
				++charnum;
			}
			outprev = c;
		}
	c = post;
	post = fgetc(in);
	} while (c != EOF);

	fseek(in, 0L, SEEK_SET);

	return charnum;
}

int ApeSuckIn(FILE *in, char *str)
{
	int outprev = '\n', c, post, i = 0, lines = 0;
	char count = YES;

	fseek(in, ftell(in), SEEK_SET);

	c = fgetc(in);
	post = fgetc(in);
	do {
		if (c == '#')
			count = NO;
		else if (c == '\n')
			count = YES;
		if (count == YES) {
			if (outprev == ' ' && c == ' ')
				;
			else if (outprev == '\n' && c == '\n')
				;
			else if (outprev == '\n' && c == ' ')
				;
			else if (c == ' ' && post == '#')
				;
			else if (c == ' ' && post == '\n')
				;
			else {
				*(str + i++) = c;
				if (c == '\n') ++lines;
			}
			outprev = c;
		}
	c = post;
	post = fgetc(in);
	} while (c != EOF);
	*(str + i) = '\0';
	return lines;
}

int NumArgs(char *strptr)
{
	int i = 0, count = 0;
	char c, prev;

	while ((prev = *(strptr + i++)) != '\0') {
		c = *(strptr + i);
		if (c == ' ' || c == '\n' || c == '\0') {
			if (prev != ' ' && prev != '\n' && prev != '\0')
				++count;
		}
	}

	return count;
}

void **ApeSetPointers(int numlines, char *str)
{
	int i = 0, arg = 0, line = 0;
	char c, prev = '\n';
	void **ptrlist;

	ptrlist = (void **) inptr.p;

	while ((c = *(str + i++)) != '\0') {
		if (prev == ' ') {
			if (c != ' ' && c != '\n' && c != '\0') {
				*(ptrlist + arg++ + numlines) = str + i - 1;
				++*(inptr.apl + line - 1);
			}
		}
		else if (prev == '\n') {
			if (c != ' ' && c != '\n' && c != '\0') {
				*(ptrlist + arg++ + numlines) = NULL;
				*(ptrlist + arg + numlines) = str + i - 1;
				*(ptrlist + line++) = (ptrlist + arg++ + numlines);
				++*(inptr.apl + line - 1);
			}
		}
		prev = c;
	}
	*(ptrlist + line) = NULL;
	*(ptrlist + arg + numlines) = NULL;
	return ptrlist;
}

void ReDoString(char *str)
{
	do {
		if (*str == ' ' || *str == '\n')
			*str = '\0';
	} while (*++str != '\0');
}

void PrintArgs(void)
{
	int i, n;

	for (i = 0; i < apein.lines; ++i) {
		printf("%d\t", i);
		for (n = 0; n < *(inptr.apl + i); ++n) {
			printf("%s ", *(*(inptr.p + i) + n));
		}
		printf("\n");
	}
}

int CheckArgs(int argc, char *argv[], char *str)
{
	int i;
	for (i = 1; i < argc; ++i)
		if (!strcmp(*(argv + i), str))
			return i;
	return 0;
}

int GetNumRuns(int lines)
{
	int i, n, runs = 0;

	for (i = 0; i < lines; ++i) {
		for (n = 0; n < *(inptr.apl + i); ++n) {
			if (!strcmp("-stim", *(*(inptr.p + i) + n))) {
				for (n = 0; n < *(inptr.apl + i); ++n) {
					if (!strcmp("-r", *(*(inptr.p + i) + n))) {
						*(inptr.stm + i) = atoi(*(*(inptr.p + i) + n + 1));
						runs += *(inptr.stm + i);
						n = *(inptr.apl + i);
					}
				}
				if (*(inptr.stm + i) == 0) {
					*(inptr.stm + i) = 1;
					runs += 1;
				}
				n = *(inptr.apl + i);
			}
		}
	}
	return runs;
}

void ApeInit(int argc, char *argv[])
{
	ApeInitArgs(argc, argv);
	ApeInitMem();
	ApeInitRuns();
	ApeInitVals("-o", defoffset, apeoffset);
	ApeInitVals("-g", defgain, apegain);
	ApeInitVals("-b", defbias, apebias);
	AddBias();
	LogVals();
	UseFeedback();
}

void AddBias(void)
{
	int i;
	for (i = 0; i < aperuns; ++i)
		*(apeoffset + i) += *(apegain + i) * *(apebias + i);
}

void SetupResponse(void)
{
	LoopButton = LoopButtonbox;
	ClearButton = ClearKeyboard;

	//LoopButton = ButtonResponse;

	ResponseSound = ResponsePC;
}

void ApeInitVals(char *id, double defval, double *doubleptr)
{
	int i, n, p;
	for (i = 0; i < aperuns; ++i)
		*(doubleptr + i) = defval;
	for (i = p = 0; i < apein.lines; ++i) {
		if (*(inptr.stm + i) > 0) {
			for (n = 0; n < *(inptr.apl + i); ++n) {
				if (!strcmp(id, *(*(inptr.p + i) + n))) {
					*(doubleptr + p) = (double) atof(*(*(inptr.p + i) + n + 1));
					break;
				}
			}
			++p;
		}
	}
}

void LogVals(void)
{
	int i;
	for (i = 0; i < aperuns; ++i) {
		*(apegain + i) = log10(1.0 + ((8.0 * *(apegain + i)) / *(apeoffset + i))) / 8.0;
		*(apeoffset + i) = log10(*(apeoffset + i));
	}
}

void UseFeedback(void)
{
	int i, n, p;
	for (i = p = 0; i < apein.lines; ++i) {
		if (*(inptr.stm + i) > 0) {
			for (n = 0; n < *(inptr.apl + i); ++n) {
				if (!strcmp("-F", *(*(inptr.p + i) + n))) {
					*(feedback + p) = YES;
					break;
				}
			}
			++p;
		}
	}
}

void ApeInitRuns(void)
{
	int i, n, p;
	if (apeinput != NULL) {
		for (i = p = 0; i < apein.lines; ++i) {
			for (n = 0; n < *(inptr.stm + i); ++n) {
				*(stimargs + p++) = GetStimPtr(i);
			}
		}
	}
	else {
		*stimargs = onestimargs;
	}
}

char **GetStimPtr(int i)
{
	int n;
	for (n = 0; n < *(inptr.apl + i); ++n) {
		if (!strcmp("-stim", *(*(inptr.p + i) + n))) {
			return (*(inptr.p + i) + n);
		}
	}
	return NULL;
}

void ApeInitArgs(int argc, char *argv[])
{
	int i;
	ApeDoArgs(argc, argv);
	if (apeinput != NULL) {
		aperuns = StrAndPtrs(apeinput);
		for (i = 0; i < apein.lines; ++i) {
			if (*(inptr.stm  + i) == 0) {
				ApeDoArgs(*(inptr.apl + i), *(inptr.p + i));
			}
		}
	}
}

void ApeDoArgs(int argc, char **argv)
{
	int i;
	for (i = 0; i < argc; ++i) {
		if (!strcmp(*(argv + i), "-r"))
			aperuns = atoi(argv[++i]);
		else if (!strcmp(*(argv + i), "-t"))
			apetrials = atoi(argv[++i]);
		else if (!strcmp(*(argv + i), "-v"))
			apeinvert = YES;
		else if (!strcmp(*(argv + i), "-q"))
			apequiet = YES;
		else if (!strcmp(*(argv + i), "-p"))
			printape = YES;
		else if (!strcmp(*(argv + i), "-N"))
			ctadapt = FALSE;

		else if (!strcmp(*(argv + i), "-i"))
			apeinput = argv[++i];
		else if (!strcmp(*(argv + i), "-f"))
			apeoutput = argv[++i];

		else if (!strcmp(*(argv + i), "-stim"))
			onestimargs = &argv[++i];
		else if (!strcmp(*(argv + i), "-init"))
			startargs = &argv[i];
		else if (!strcmp(*(argv + i), "-end"))
			endargs = &argv[i];

		else if (!strcmp(*(argv + i), "-o"))
			defoffset = atof(argv[++i]);
		else if (!strcmp(*(argv + i), "-g"))
			defgain = atof(argv[++i]);


		else if (!strcmp(*(argv + i), "-L"))
			apelevels = atoi(argv[++i]);
		else if (!strcmp(*(argv + i), "-G"))
			dragtrials = atoi(argv[++i]);

		else if (!strcmp(*(argv + i), "-h") || !strcmp(*(argv + i), "-?")) {
			PrintInfo();
			exit(1);
		}
	}
}

void ApeInitMem(void)
{
	trial = (int *) ApeCheckMem(malloc(sizeof(int) * aperuns));
	stimulus = (int *) ApeCheckMem(malloc(sizeof(int) * aperuns));

	apebias = (double *) ApeCheckMem(malloc(sizeof(double) * aperuns));
	apeoffset = (double *) ApeCheckMem(malloc(sizeof(double) * aperuns));
	apegain = (double *) ApeCheckMem(malloc(sizeof(double) * aperuns));
	ape_sd = (double *) ApeCheckMem(malloc(sizeof(double) * aperuns));
	ape_ed50 = (double *) ApeCheckMem(malloc(sizeof(double) * aperuns));
	Csd = (double *) ApeCheckMem(malloc(sizeof(double) * aperuns));
	Ced50 = (double *) ApeCheckMem(malloc(sizeof(double) * aperuns));

	srlist.stims = (int *) ApeCheckMem(malloc(sizeof(int) * apetrials * aperuns));
	srlist.resps = (int *) ApeCheckMem(malloc(sizeof(int) * apetrials * aperuns));

	srtable.stim = (int *) ApeCheckMem(malloc(sizeof(int) * apelevels));
	srtable.test = (int *) ApeCheckMem(malloc(sizeof(int) * apelevels));
	srtable.resp = (int *) ApeCheckMem(malloc(sizeof(int) * apelevels));
	srtable.prop = (double *) ApeCheckMem(malloc(sizeof(double) * apelevels));
	srtable.prob = (double *) ApeCheckMem(malloc(sizeof(double) * apelevels));
	srtable.weight = (double *) ApeCheckMem(malloc(sizeof(double) * apelevels));

	stimargs = (char ***) ApeCheckMem(malloc(sizeof(void *) * aperuns));

	feedback = (boolean *) ApeCheckMem(malloc(sizeof(boolean) * aperuns));

	SetMemToZero();
}

void SetMemToZero(void)
{
	int i;

	for (i = 0; i < aperuns; ++i) {
		*(trial + i) = *(stimulus + i) = 0;
		*(apeoffset + i) = *(apegain + i) = *(ape_sd + i) = *(apebias + i) = 0;
		*(ape_ed50 + i) = *(Csd + i) = *(Ced50 + i) = 0;
		*(feedback + i) = NO;
	}
	for (i = 0; i < (apetrials * aperuns); ++i) {
		*(srlist.stims + i) = *(srlist.resps + i) = 0;
	}
	for (i = 0; i < apelevels; ++i) {
		*(srtable.stim + i) = *(srtable.test + i) = *(srtable.resp + i) = 0;
		*(srtable.prop + i) = *(srtable.prob + i) = *(srtable.weight + i) = 0.0;
	}
}

void ApeClearMem(void)
{
	free(trial);
	free(stimulus);
	free(apebias);
	free(apeoffset);
	free(apegain);
	free(ape_sd);
	free(ape_ed50);
	free(Csd);
	free(Ced50);
	free(srlist.stims);
	free(srlist.resps);
	free(srtable.stim);
	free(srtable.test);
	free(srtable.resp);
	free(srtable.prop);
	free(srtable.prob);
	free(srtable.weight);
	free(inptr.apl);
	free(inptr.stm);
	free(inptr.p);
	free(apein.ptr);
	free(stimargs);
	free(feedback);
}

void *ApeCheckMem(void *ptr)
{
	if (ptr == NULL) {
		fprintf(stderr, "Not enough memory to initialise variables\n");
		exit(1);
	}
	return ptr;
}

char *dtoa(double num, char *str, int length)
{
	short int i = 0, p = 0, n = 0, len, offset = 0;
	char *tmp;

	if (num == 0.0) {
		*(str + p++) = '0';
		*(str + p++) = '\0';
		return str;
	}

	if (num < 0.0) {
		num *= -1.0;
		offset = 1;
		*str = '-';
	}

	tmp = (char *) malloc(sizeof(char) * length);

	while (num <= MAXLONG/10) {
		num *= 10;
		++i;
	}

	len = strlen(ltoa((unsigned long int) num, tmp, 10));

	if (i >= len) {
		*(str + offset + p++) = '0';
		*(str + offset + p++) = '.';
		while (i-- > len)
			*(str + offset + p++) = '0';
		for (i = 0; i <= len; ++i)
			*(str + offset + p++) = *(tmp + i);
	}

	else {
		while (p < len - i) {
			*(str + p++ + offset) = *(tmp + n++);
		}
		*(str + p++ + offset) = '.';
		while (n <= len) {
			*(str + p++ + offset) = *(tmp + n++);
		}
	}

	while (*(str + offset + len--) == '0')
		;
	*(str + len + offset + 2) = '\0';
	return str;
}

char *DoApeParameters(double dnum, boolean invt, char *str, int length)
{
	int len;

	len = strlen(dtoa(dnum, str, length));

	if (len > length - 2)
		return NULL;

	*(str + len++) = ' ';
	*(str + len++) = (invt == 1) ? '1' : '0';
	*(str + len) = '\0';
	return str;
}

int DisplayStimulus(double parameter, double baseval)
{
	boolean invert = 0;

	if (parameter < baseval)
		expected = A_UP;
	else if (parameter > baseval)
		expected = C_UP;
	else
		expected = EQUAL_TO;

	if (apeinvert == YES)
		invert = (rand() > (RAND_MAX / 2)) ? 0 : 1;

	(*runf)(invert, pow(10, parameter), *(stimargs + run));

	(*ClearButton)();

	return invert;
}

void RunCommandCom(void)
{
	if (spawnlp(P_WAIT, "command.com", NULL) == -1) {
		perror("cannot spawn command.com");
		exit(1);
	}
}

void PrintParameters(void)
{
	int i;
	FILE *out = stdout;

	printf("\noutput: %s\tinvert: %s\n", apeoutput, (apeinvert == YES) ? "on" : "off");
	printf("dragtrials: %d\t\taperuns:    %d\t\tapetrials: %d\n\n", dragtrials, aperuns, apetrials);

	for (i = 0; i < aperuns; ++i) {
		printf("run %d:\t gain = %f (%f) offset = %f (%f)\n", (i + 1), *(apegain + i), pow(10, *(apegain + i)), *(apeoffset + i), pow(10, *(apeoffset + i)));
		PrintArgv(*(stimargs + i), out);
	}
}

void PrintArgv(char *argv[], FILE *out)
{
	int i = 0;
	while (*(argv + i) != NULL)
		fprintf(out, " %s", *(argv + i++));
	fprintf(out, "\n");
}

void PrintInfo(void)
{
	printf("  -i     <input file>  -f     <output file>  [output.ape]\n");
	printf("  -stim  <stimulus executable>\n");
	printf("  -init  <initialising executable>\n");
	printf("  -end   <closing executable>\n");
	printf("  -r <runs>    [1]     -o <offset> [0.5]  -g <gain>         [0.01]\n");
	printf("  -t <trials>  [64]    -L <levels> [32]   -G <dragtrials>   [16]\n");
	printf("  -D <density> [4]     -R <range>  [3]\n");
	printf("  -v {invert}  [NO]    -q {quiet}  [NO]   -p {print values} [NO]\n");
	printf("  <> must put value after flag   [] default value  {} description\n\n");

	printf("# A SAMPLE INPUT FILE\n");
	printf("# NOTE - only offset, gain and runs can be set on stimulus lines. If these are\n");
	printf("# set on a non stimulus line then these will form the default. A stimulus line\n");
	printf("# is any line with a \"-stim\" on it. Any variables set on these lines will over-\n");
	printf("# ride the default values but only for that particular stimulus\n");
	printf("# Next line sets default offset and gain, establishes output file etc...\n");
	printf("-t 64 -v -q -f contrast.out -o 0.5 -g 0.01     # non stimulus line\n");
	printf("-init vsginit.exe                              # stimulus initialisation\n");
	printf("-o 0.60 -stim contrast.exe -sf 5.0 -tf 6.0     # first stimulus line\n");
	printf("-o 0.45 -g 0.01 -r 2 -stim contrast.exe -sf 4.0 -tf 6.0\n");
	printf("-o 0.55 -r 2 -stim contrast.exe -sf 3.0 -tf 6.0\n");
	printf("-end vsgclose.exe                              # blanks screen or whatever\n");
}

int GetStimVal(double offset, double sd)
{
	int rnd_num;

	rnd_num = (int) ((((double) rand()) / ((double) RAND_MAX)) * 4.0);

	if (rnd_num == 0)
		return dtoi(offset + ((APEWIDTH / 2.0) * sd));
	else if (rnd_num == 1)
		return dtoi(offset - ((APEWIDTH / 2.0) * sd));
	else if (rnd_num == 2)
		return dtoi(offset + ((APEWIDTH / 6.0) * sd));
	else if (rnd_num == 3)
		return dtoi(offset - ((APEWIDTH / 6.0) * sd));
	else
		return GetStimVal(offset, sd);
}
