#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <math.h>
#include <time.h>
#include <direct.h>
#include <conio.h>
#include <iostream.h>

//#include "vsgex2.h"
#include <VSGV8.h>

#define WIN32_LEAN_AND_MEAN		// Exclude rarely-used stuff from Windows headers

#define LUMRATIO 1 //(23.535)
//#define LUMRATIO 28.46/42.2 //(58.3 / 67) //first value should be the lowest monitor luminance when LUMRATIO = 1
							 //Bristol = 58.3 (80Hz); Belfast = 36.01 (120Hz); Sydney = 33.6 (120Hz)

#define NUMDOTS 512		// the number of dots needs to be divisible by 4
#define NUMTESTDOTS 512 // the number of dots needs to be divisible by 4
#define DOTSIZE 2  
#define IMSIZE 201
#define PI 3.1415927
#define PIXELS_PER_CM 25.8 
#define FRAMERATE 120
#define VIEWDIST 71

#define COHERENCE 1.0

#define SPEED 5.0
#define YES 1
#define NO 0
 
// TIMING PARAMS
#define ADAPT_TIME 30.0 //30.0 // adaptation time in seconds
#define TOPUP_TIME 5.0 //5.0       // topup time in seconds
//#define TEST_TIME 0.5 //0.2     // time that the match stimulus is up for
#define STIM_PAUSE 0.0     // time between stimuli - note that isi is specified on the command line

#define TIMER_FREQ 1193180L   /* Timer freq = 1.99 MHz */
#define TIMER_COUNT 0x42      /* 8253 timer count */
#define TIMER_MODE 0x43       /* 8253 timer control port */
#define TIMER_OSC 0xb6        /* To use timer as oscilator */
#define OUT_8255 0x61         /* 8255 PPI output port adrs */
#define SPKRON 3              /* Bit 0 = control spkr by timer */
                              /* Bit 1 = speaker on/off */


CBBOXRECORD CB6Box;  // Call CB6 response box.  Also needs to be called in logape.c


typedef struct {
	double adapt_coherence;
	double adapt_speed;
	double adapt_angle;
	double test_coherence;
	double test_speed_1;
	double test_angle_1;
	double test_speed_2;
	double test_angle_2;
	double test_time_1;
	double test_time_2;
	int isi; // time between adapt and match stimulus - msec
	char oscillate;
} STIMULUS;

extern int ApeMain(int argc, char *argv[], void *initfunc, void *runfunc, void *endfunc);

void RunDots(double displaytime, double coherence, double speed, double angle, char oscillate);
void RunTestDots(double displaytime, double speed, double angle);
double drand(void);
int bound(double n);
void endfunc(void);
STIMULUS *GetArgs(int invert, double apeval, char *argv[]);
void initfunc(void);
void runfunc(int invert, double apeval, char *argv[]);
void cpbSleep(long wait_time);


double CoordBuff[NUMDOTS][2];
int DotPositions[NUMDOTS][2]; 
double CoordBuff2[NUMTESTDOTS][2];
int DotPositions2[NUMTESTDOTS][2];

static char firsttrial = YES;

int main(int argc, char *argv[])
{
	srand(time(NULL));

	//initfunc();
	//runfunc(0, 0, argv);
	//endfunc();

	// if you want to run ape then just uncoment the next line and comment out the
	// three lines above.
	return ApeMain(argc, argv, (void *) initfunc, (void *) runfunc, (void *) endfunc);
}

void initfunc(void)
{
	int CheckCard, i;
	VSGLUTBUFFER buff;
	char x = 88;


	//Initialise the vsg card then check that it initialised O.K.
	CheckCard = vsgInit("");
	if (CheckCard < 0)
		_exit(0);

	//Set the draw mode to use (enable antialiasing).
	vsgSetPenSize(2,2);
	vsgSetDrawMode(vsgCENTREXY + vsgSOLIDPEN);

	//Load the true type font that you want to write with.
	//Note: the font name should be used and NOT the font filename.
	vsgSetTrueTypeFont("Arial");

	//Set the draw modes for the text i.e. centred, italic, angle, size.
	vsgSetStringMode(0, 20, vsgALIGNCENTRETEXT, vsgALIGNCENTRETEXT, 0, 0);

	//cbboxOpen(respCB3);  //open communication with CB3 box
	cbboxOpen(respCB6);  //open communication with CB6 box

	//Clear a page in memory and display it.
	vsgSetDrawPage(vsgVIDEOPAGE, 0, 128);

	//Display the cleared memory page.
	vsgSetDisplayPage(0);

	// create and load a greyscale palette
	for (i =0; i < 256; ++i) {
		buff[i].a = buff[i].b = buff[i].c = LUMRATIO * (double) i / (double) 255;
	}
	vsgLUTBUFFERWrite(0, (VSGLUTBUFFER *) & buff);
	vsgLUTBUFFERtoPalette(0);

	//Initialise the random dot system.
	vsgDOTInitialise((NUMDOTS + NUMTESTDOTS) * 2, 3, vsgDOTAUTORESERVE);

	//vsgSetCommand(vsgOVERLAYMASKMODE);
	vsgSetDrawPage(vsgOVERLAYPAGE, 0, 1);
	vsgPaletteWriteOverlayCols((VSGLUTBUFFER *) &buff[128], 1, 1);
	vsgPaletteWriteOverlayCols((VSGLUTBUFFER *) &buff[255], 2, 1);
	vsgPaletteWriteOverlayCols((VSGLUTBUFFER *) &buff[255], 3, 1);
	vsgSetDrawOrigin(vsgGetScreenWidthPixels() / 2, vsgGetScreenHeightPixels() / 2);

	// draw windows to display stimuli through + fixation spot
	for (i = 0; i < 10; ++i) {
		vsgSetPen1(0);
		vsgDrawOval(0, 0, IMSIZE, IMSIZE);
	}

	vsgSetDrawPage(vsgVIDEOPAGE, 0, 128);
	vsgSetDrawPage(vsgVIDEOPAGE, 1, 128);
	vsgSetDrawPage(vsgVIDEOPAGE, 2, 128);

	vsgSetPen1(0);
	vsgSetPen2(128);

	//Change the drawing origin.
	vsgSetDrawOrigin(0, 0);

	//Specify the VIDEO page to draw the dots onto.
	vsgSetDrawPage(vsgVIDEOPAGE, 1, 128);

	//Draw some text.
	vsgDrawChar(512, 385, x);
	vsgSetZoneDisplayPage(vsgVIDEOPAGE, 1);


	//Load the coordinates buffers with a random set of coordinates.
	for (i = 0 ; i < NUMDOTS ; i++ ) 
	{
		CoordBuff[i][0] = drand() * (double) IMSIZE;
		CoordBuff[i][1] = drand() * (double) IMSIZE;

	}

	for (i = 0 ; i < NUMTESTDOTS ; i++ ) {
		CoordBuff2[i][0] = drand() * (double) IMSIZE;
		CoordBuff2[i][1] = drand() * (double) IMSIZE;
	}

	while (cbboxCheck((CBBOXRECORD*)&CB6Box)==respEMPTY) // Button press starts run
		;
	cbboxFlush();
}

void runfunc(int invert, double apeval, char *argv[])
{
	STIMULUS *stim;
	double adapttime, random_sequence;
	int i;
	char x = 88;

	random_sequence = drand(); //used to randomise presentatin order of test stimuli

	if (firsttrial == YES) {
		adapttime = ADAPT_TIME;
	}
	else {
		adapttime = TOPUP_TIME;
	}

	stim = GetArgs(invert, apeval, argv);


	// do stim pause
	//cpbSleep((long) (STIM_PAUSE * 1000000));


		// run adaptation sequence

		RunDots(adapttime, stim->adapt_coherence, stim->adapt_speed, stim->adapt_angle, stim->oscillate);


		//Specify the VIDEO page to draw X onto.
		vsgSetDrawPage(vsgVIDEOPAGE, 1, 128);

		//Draw X
		//Set the draw modes for the text i.e. centred, italic, angle, size.
		vsgSetStringMode(0, 20, vsgALIGNCENTRETEXT, vsgALIGNCENTRETEXT, 0, 0);
		vsgDrawChar(512, 385, x);
		vsgSetZoneDisplayPage(vsgVIDEOPAGE, 1);

		// do ISI - specified in msec
		cpbSleep((long) stim->isi * 1000L);


		// run test sequence 1
		if(random_sequence < 0.5)
			RunDots(stim->test_time_1, stim->test_coherence, stim->test_speed_1, stim->test_angle_1, NO);
		else
			RunTestDots(stim->test_time_2, stim->test_speed_2, stim->test_angle_2);


		//Specify the VIDEO page to draw X onto.
		vsgSetDrawPage(vsgVIDEOPAGE, 1, 128);

		//Draw X
		//Set the draw modes for the text i.e. centred, italic, angle, size.
		vsgSetStringMode(0, 20, vsgALIGNCENTRETEXT, vsgALIGNCENTRETEXT, 0, 0);
		vsgDrawChar(512, 385, x);
		vsgSetZoneDisplayPage(vsgVIDEOPAGE, 1);


		// do ISI - specified in msec
		cpbSleep((long) stim->isi * 1000L);


		// run test sequence 2
		if(random_sequence < 0.5)
			RunTestDots(stim->test_time_2, stim->test_speed_2, stim->test_angle_2);
		else
			RunDots(stim->test_time_1, stim->test_coherence, stim->test_speed_1, stim->test_angle_1, NO);


		//Specify the VIDEO page to draw X onto.
		vsgSetDrawPage(vsgVIDEOPAGE, 1, 128);

		//Draw X
		//Set the draw modes for the text i.e. centred, italic, angle, size.
		vsgSetStringMode(0, 20, vsgALIGNCENTRETEXT, vsgALIGNCENTRETEXT, 0, 0);
		vsgDrawChar(512, 385, x);
		vsgSetZoneDisplayPage(vsgVIDEOPAGE, 1);

		
		//Load the coordinates buffers with a random set of coordinates.
	for (i = 0 ; i < NUMDOTS ; i++ ) 
	{
		CoordBuff[i][0] = drand() * (double) IMSIZE;
		CoordBuff[i][1] = drand() * (double) IMSIZE;
	}

	for (i = 0 ; i < NUMTESTDOTS ; i++ ) 
	{
		CoordBuff2[i][0] = drand() * (double) IMSIZE;
		CoordBuff2[i][1] = drand() * (double) IMSIZE;
	}
}

void RunDots(double displaytime, double coherence, double speed, double angle, char oscillate)
{
	int i;
	long f, numframes, switch_dir = 600 * 1000L;
	int XOffset, YOffset;
	double ppd, displace;
	double xshift, yshift, shift;
	int page;
	int *dotptr;
	double *coordptr;
	int radius = IMSIZE / 2, a, b, rad_x, rad_y;
	char x = 88;

	// clear all three pages
	vsgSetDrawPage(vsgVIDEOPAGE, 0, 128);
	vsgSetDrawPage(vsgVIDEOPAGE, 1, 128);

	// calculate the number of frames
	numframes = (long) floor((displaytime * vsgGetSystemAttribute(vsgFRAMERATE)) + 0.5);


	ppd = PIXELS_PER_CM * VIEWDIST * (double) (2.0 * tan(PI / 360.0)); // pixels per degree

	displace = 3.7 * ppd;

	//Change the drawing origin.
	vsgSetDrawOrigin(0, 0);

	//Find out the horizontal resolution of the vsg screen.
	YOffset = div((vsgGetScreenHeightPixels()-IMSIZE), 2).quot;
	XOffset = div((vsgGetScreenWidthPixels()-(IMSIZE*2 + (int)(displace))), 2).quot;
	a = XOffset;
	b = YOffset;

	//cycle through loop changing each dots coordinates slightly then redrawing the
	//dots onto a different page.

	vsgSetZoneDisplayPage(vsgVIDEOPAGE, 0);

	page = 1;

	for (f = 0; f < numframes; ++f) {

		coordptr = &CoordBuff[0][0];

		shift = ppd * speed / FRAMERATE; // pixels per frame

		for (i = 0; i < NUMDOTS; ++i) {

			if(drand() <= coherence) {
				xshift = cos(angle * PI / 180) * shift;
				yshift = sin(angle * PI / 180) * shift;
				*coordptr++ += xshift;				
				*coordptr++ += yshift;
			}

			else{
				xshift = cos((drand() * (double) 360) * PI / 180) * shift;
				yshift = sin((drand() * (double) 360) * PI / 180) * shift;
				*coordptr++ += xshift;				
				*coordptr++ += yshift;
			}
		}


		// wrap around
		coordptr = &CoordBuff[0][0];
		for (i = 0; i < NUMDOTS * 2; ++i) {
			while (*coordptr > IMSIZE)
				*coordptr -= IMSIZE;
			while (*coordptr < 0)
				*coordptr += IMSIZE;
			++coordptr;
		}

		// get nearest positions
		coordptr = &CoordBuff[0][0];
		dotptr = &DotPositions[0][0];
		for (i = 0; i < NUMDOTS; ++i) {
			*dotptr++ = ((int) *coordptr++) + a;
			*dotptr++ = ((int) *coordptr++) + b;
		}
		dotptr = &DotPositions[0][0];


		// load the dots (black & white)
		vsgDOTInitLoadGroup(vsgDOTXYLEVEL, DOTSIZE, DOTSIZE, 0);
		for (i = 0 ; i < NUMDOTS; i++) {
			rad_x = *dotptr - (IMSIZE / 2) - a;
			rad_y = *(dotptr + 1) - (IMSIZE / 2) - b;
			if(sqrt((rad_x * rad_x) + (rad_y * rad_y)) < radius) 
			{
			if(i%2 == 1)
			vsgDOTGenerateNormalDot(*dotptr, *(dotptr + 1), 255);
			else
			vsgDOTGenerateNormalDot(*dotptr, *(dotptr + 1), 0);
			}
			dotptr += 2;
		}
		vsgDOTLoadGroup();

		//Specify the VIDEO page to draw the dots onto.
		vsgSetDrawPage(vsgVIDEOPAGE, page, 128);

		//Draw some text.
		vsgDrawChar(512, 385, x);

		//Draw and display the dots.
		vsgDrawDots(vsgDOTNOMAKEVISIBLE);
		vsgSetZoneDisplayPage(vsgVIDEOPAGE, page);

		//Toggle the value in the page parameter.
		page = (page == 0) ? 1 : 0;
	}
	// clear all pages
	vsgSetDrawPage(vsgVIDEOPAGE, 0, 128);
	vsgSetDrawPage(vsgVIDEOPAGE, 1, 128);
	vsgSetDrawPage(vsgVIDEOPAGE, 2, 128);

	if (firsttrial == YES) firsttrial = NO;
}


void RunTestDots(double displaytime, double speed, double angle)
{
	int i;
	long f, numframes;
	int XOffset, YOffset;
	double ppd, displace;
	double xshift, yshift, shift;
	int page;
	int *dotptr;
	double *coordptr;
	int radius = IMSIZE / 2, a, b, rad_x, rad_y;
	char x = 88;

	// clear all three pages
	vsgSetDrawPage(vsgVIDEOPAGE, 0, 128);
	vsgSetDrawPage(vsgVIDEOPAGE, 1, 128);

	// calculate the number of frames
	numframes = (long) floor((displaytime * vsgGetSystemAttribute(vsgFRAMERATE)) + 0.5);


	ppd = PIXELS_PER_CM * VIEWDIST * (double) (2.0 * tan(PI / 360.0)); // pixels per degree

	displace = 3.7 * ppd;

	//Change the drawing origin.
	vsgSetDrawOrigin(0, 0);

	//Find out the horizontal resolution of the vsg screen.
	YOffset = div((vsgGetScreenHeightPixels()-IMSIZE), 2).quot;
	XOffset = div((vsgGetScreenWidthPixels() + (int)(displace)), 2).quot;
	a = XOffset;
	b = YOffset;


	//cycle through loop changing each dots coordinates slightly then redrawing the
	//dots onto a different page.

	
	shift = ppd * speed / FRAMERATE; // pixels per frame
	xshift = cos(angle * PI / 180) * shift;
	yshift = sin(angle * PI / 180) * shift;



	vsgSetZoneDisplayPage(vsgVIDEOPAGE, 0);

	page = 1;

	for (f = 0; f < numframes; ++f) {

		coordptr = &CoordBuff2[0][0];
		for (i = 0; i < NUMTESTDOTS; ++i) {
				*coordptr++ += xshift;
				*coordptr++ += yshift;

			shift = ppd * speed / FRAMERATE; // pixels per frame
			xshift = cos(angle * PI / 180) * shift;
			yshift = sin(angle * PI / 180) * shift;
		}


		// rap around
		coordptr = &CoordBuff2[0][0];
		for (i = 0; i < NUMTESTDOTS * 2; ++i) {
			while (*coordptr > IMSIZE)
				*coordptr -= IMSIZE;
			while (*coordptr < 0)
				*coordptr += IMSIZE;
			++coordptr;
		}

		// get nearest positions
		coordptr = &CoordBuff2[0][0];
		dotptr = &DotPositions2[0][0];
		for (i = 0; i < NUMTESTDOTS; ++i) {
			*dotptr++ = ((int) *coordptr++) + a;
			*dotptr++ = ((int) *coordptr++) + b;
		}
		dotptr = &DotPositions2[0][0];




		// load the dots (black & white)
		vsgDOTInitLoadGroup(vsgDOTXYLEVEL, DOTSIZE, DOTSIZE, 0);
		for (i = 0 ; i < NUMDOTS; i++) {
			rad_x = *dotptr - (IMSIZE / 2) - a;
			rad_y = *(dotptr + 1) - (IMSIZE / 2) - b;
			if(sqrt((rad_x * rad_x) + (rad_y * rad_y)) < radius) 
			{
			if(i%2 == 1)
			vsgDOTGenerateNormalDot(*dotptr, *(dotptr + 1), 255);
			else
			vsgDOTGenerateNormalDot(*dotptr, *(dotptr + 1), 0);
			}
			dotptr += 2;
		}
		vsgDOTLoadGroup();



		//Specify the VIDEO page to draw the dots onto.
		vsgSetDrawPage(vsgVIDEOPAGE, page, 128);

		//Draw some text.
		vsgDrawChar(512, 385, x);

		//Draw and display the dots.
		vsgDrawDots(vsgDOTNOMAKEVISIBLE);
		vsgSetZoneDisplayPage(vsgVIDEOPAGE, page);

		//Toggle the value in the page parameter.
		page = (page == 0) ? 1 : 0;


	}
	// clear all pages
	vsgSetDrawPage(vsgVIDEOPAGE, 0, 128);
	vsgSetDrawPage(vsgVIDEOPAGE, 1, 128);
	vsgSetDrawPage(vsgVIDEOPAGE, 2, 128);


}


double drand(void)
{
	return (double) rand() / (double) RAND_MAX;
}

int nearest(double n)
{
	return (n >= 0) ? (int) floor(n + 0.5) : (int) ceil(n - 0.5);
}

int bound(double n)
{
	while (n >= IMSIZE)
		n -= IMSIZE;

	while (n < 0)
		n += IMSIZE;

	return (int) floor(n);
}



STIMULUS *GetArgs(int invert, double apeval, char *argv[])
{
	static STIMULUS stim;

	// set stimulus defaults
	stim.test_coherence = COHERENCE;
	stim.adapt_coherence = COHERENCE;
	stim.adapt_speed = SPEED;
	stim.test_speed_1 = SPEED;
	stim.test_speed_2 = SPEED;
	stim.adapt_angle = 90;
	stim.test_angle_1 = 180;
	stim.test_angle_2 = 180;
	stim.test_time_1 = 0.6;
	stim.test_time_2 = 0.6;
	stim.isi = 500;
	stim.oscillate = NO;

	while (*argv != NULL) {

		if (!strcmp("-tc", *argv))
			stim.test_coherence = atof(*(argv + 1));

		else if (!strcmp("-ac", *argv))
			stim.adapt_coherence = atof(*(argv + 1));

		else if (!strcmp("-ts1", *argv))
			stim.test_speed_1 = atof(*(argv + 1));

		else if (!strcmp("-ts2", *argv))
			stim.test_speed_2 = atof(*(argv + 1));

		else if (!strcmp("-as", *argv))
			stim.adapt_speed = atof(*(argv + 1));

		else if (!strcmp("-ta1", *argv))
			stim.test_angle_1 = atof(*(argv + 1));

		else if (!strcmp("-ta2", *argv))
			stim.test_angle_2 = atof(*(argv + 1));

		else if (!strcmp("-aa", *argv))
			stim.adapt_angle = atof(*(argv + 1));

		else if (!strcmp("-tt1", *argv))
			stim.test_time_1 = atof(*(argv + 1));

		else if (!strcmp("-tt2", *argv))
			stim.test_time_2 = atof(*(argv + 1));

		else if (!strcmp("-isi", *argv))
			stim.isi = atoi(*(argv + 1));

		else if (!strcmp("-OSCILLATE", *argv))
			stim.oscillate = YES;

		++argv;
	}

		stim.adapt_angle -= 2 * stim.adapt_angle;
		stim.test_angle_1 -= 2 * stim.test_angle_1;
		stim.test_angle_2 -= 2 * stim.test_angle_2;

	// remember use log ape so can never have a zero or negative apeval
	if (apeval > 0)
		stim.test_time_2 = apeval;

	return &stim;
}

void endfunc(void)
{
	printf("Finished\n");
	cbboxClose();
}

void cpbSleep(long wait_time)
{
	vsgResetTimer();
	while (vsgGetTimer() < wait_time)
		;
}
